from yalc.clients.client import Client
from yalc.clients.provider_clients.anthropic import AnthropicClient
from yalc.clients.provider_clients.openai import OpenAIClient
from yalc.clients.strategy import ClientLogStrategy
from yalc.common.schemas import LLMModel, LLMProvider

provider_to_client_map: dict[LLMProvider, type[Client]] = {
    LLMProvider.OPENAI: OpenAIClient,
    LLMProvider.ANTHROPIC: AnthropicClient,
}


def create_client(
    model: LLMModel,
    log_strategies: list[ClientLogStrategy] = [],
) -> Client:
    client_class = provider_to_client_map.get(model.provider)
    if client_class is None:
        raise ValueError(f"Unsupported provider: {model.provider}")
    return client_class(model, log_strategies)



from pydantic import BaseModel

from yalc.common.schemas import ContextMessage, LLMRole, ResponseStats


class ClientMessage(BaseModel):
    response: BaseModel
    role: LLMRole


class ClientCall(ResponseStats):
    context_messages: list[ContextMessage]
    client_message: ClientMessage
    model_name: str



from abc import ABC, abstractmethod

from pydantic import BaseModel

from yalc.clients.schemas import ClientCall


class ClientLogStrategy[T: BaseModel](ABC):
    @abstractmethod
    def handle(self, call: ClientCall, context: T | None):
        pass



from openai.types.responses import Response
from pydantic import BaseModel

from yalc.clients.client import Client
from yalc.common.schemas import ResponseStats


class OpenAIClient(Client):
    async def _response[T: BaseModel](
        self,
        response_type: type[T],
        messages: list[dict[str, str]],
    ) -> tuple[T, Response]:
        return await self.instructor_client.responses.create_with_completion(  # type: ignore
            model=self.model.value,
            response_model=response_type,
            max_retries=3,
            input=messages,  # type: ignore
        )

    def _get_response_stats(
        self, response: Response
    ) -> ResponseStats:
        if response.usage is None:
            raise RuntimeError("no usage for llm call")

        return self.pricing_service.build_response_stats(
            input_tokens=response.usage.input_tokens,
            output_tokens=response.usage.output_tokens,
            model=self.model,
        )



from yalc.clients.client import Client
from yalc.clients.client_factory import create_client
from yalc.clients.schemas import ClientCall, ClientMessage
from yalc.clients.strategy import ClientLogStrategy
from yalc.common.schemas import (
    ContextMessage,
    LLMModel,
    LLMProvider,
    LLMRole,
    ResponseStats,
)

__all__ = [
    "Client",
    "ClientLogStrategy",
    "ClientCall",
    "ClientMessage",
    "LLMRole",
    "ContextMessage",
    "ResponseStats",
    "create_client",
    "LLMModel",
    "LLMProvider",
]



from abc import ABC, abstractmethod
from typing import Any, overload

import instructor
from pydantic import BaseModel

from yalc.clients.schemas import ClientCall, ClientMessage
from yalc.clients.strategy import ClientLogStrategy
from yalc.common.pricing import PricingService
from yalc.common.schemas import LLMModel, LLMRole, ResponseStats
from yalc.common.utils import to_context_messages


class Client(ABC):
    def __init__(
        self,
        model: LLMModel,
        log_strategies: list[ClientLogStrategy] = [],
    ):
        self.log_strategies = log_strategies
        self.pricing_service = PricingService()

        self.instructor_client = instructor.from_provider(
            model.provider_string,
            async_client=True,
            mode=model.mode,
        )
        self.model = model

    @overload
    async def structured_response[T: BaseModel](
        self,
        response_type: type[T],
        messages: list[dict[str, str]],
        context: BaseModel,
    ) -> T: ...

    @overload
    async def structured_response[T: BaseModel](
        self,
        response_type: type[T],
        messages: list[dict[str, str]],
    ) -> tuple[T, BaseModel]: ...

    async def structured_response[T: BaseModel](
        self,
        response_type: type[T],
        messages: list[dict[str, str]],
        context: BaseModel | None = None,
    ) -> T | tuple[T, BaseModel]:
        """
        Provides a structured response via an LLM call

        Uses provided log strategies to log all the LLM messages
        """
        response, llm_call = await self._structured_response(
            response_type, messages
        )

        if context is not None:
            for strategy in self.log_strategies:
                strategy.handle(llm_call, context)
            return response

        return response, llm_call

    async def _structured_response[T: BaseModel](
        self,
        response_type: type[T],
        messages: list[dict[str, str]],
    ) -> tuple[T, ClientCall]:
        parsed, response = await self._response(
            response_type, messages
        )

        llm_call = self._create_llm_call(messages, parsed, response)
        return parsed, llm_call

    def _create_llm_call(
        self,
        messages: list[dict[str, str]],
        parsed: BaseModel,
        response: Any,
    ) -> ClientCall:
        context_messages = to_context_messages(messages)
        client_message = self._parse_client_message(parsed)

        return ClientCall(
            context_messages=context_messages,
            client_message=client_message,
            model_name=self.model,
            **self._get_response_stats(response).model_dump(),
        )

    @staticmethod
    def _parse_client_message(parsed: BaseModel) -> ClientMessage:
        return ClientMessage(response=parsed, role=LLMRole.ASSISTANT)

    @abstractmethod
    async def _response[T: BaseModel](
        self,
        response_type: type[T],
        messages: list[dict[str, str]],
    ) -> tuple[T, Any]:
        pass

    @abstractmethod
    def _get_response_stats(self, response: Any) -> ResponseStats:
        pass



from cachetools import TTLCache, cached
from litellm import model_cost_map_url
from litellm.litellm_core_utils.get_model_cost_map import (
    get_model_cost_map,
)

from yalc.common.schemas import LLMModel, ResponseStats, TokensPricing


class PricingService:
    def build_response_stats(
        self,
        input_tokens: int,
        output_tokens: int,
        model: LLMModel,
    ) -> ResponseStats:
        return ResponseStats(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            input_tokens_cost=self._get_input_token_cost(
                model.value, input_tokens
            ),
            output_tokens_cost=self.get_output_token_cost(
                model.value, output_tokens
            ),
        )

    def _get_input_token_cost(
        self, model: str, input_tokens: int
    ) -> float:
        model_pricing = self._get_model_pricing(model)
        return input_tokens * model_pricing.input_cost_per_token

    def get_output_token_cost(
        self, model: str, output_tokens: int
    ) -> float:
        model_pricing = self._get_model_pricing(model)
        return output_tokens * model_pricing.output_cost_per_token

    @staticmethod
    @cached(cache=TTLCache(maxsize=100, ttl=60 * 5))  # 5 minute TTL
    def _get_model_pricing(model: str) -> TokensPricing:
        """
        Cache the result for 5 minutes since every function call makes
        a http request
        """
        model_cost = get_model_cost_map(model_cost_map_url)
        if model not in model_cost:
            raise ValueError(
                f"Model {model} not found in litellm pricing data"
            )

        pricing = model_cost[model]
        return TokensPricing(
            input_cost_per_token=pricing.get(
                "input_cost_per_token", 0
            ),
            output_cost_per_token=pricing.get(
                "output_cost_per_token", 0
            ),
        )



from yalc.common.schemas import ContextMessage, LLMRole


def to_context_messages(messages: list[dict]) -> list[ContextMessage]:
    return [
        ContextMessage(message=m["content"], role=LLMRole(m["role"]))
        for m in messages
    ]



from enum import StrEnum
from typing import NamedTuple

import instructor
from pydantic import BaseModel


class LLMRole(StrEnum):
    ASSISTANT = "assistant"
    USER = "user"
    SYSTEM = "system"
    TOOL = "tool"


class ResponseStats(BaseModel):
    input_tokens: int
    output_tokens: int
    input_tokens_cost: float
    output_tokens_cost: float


class ContextMessage(BaseModel):
    message: str
    role: LLMRole


class TokensPricing(NamedTuple):
    input_cost_per_token: float
    output_cost_per_token: float


class LLMProvider(StrEnum):
    OPENAI = "openai"
    ANTHROPIC = "anthropic"


class LLMModel(StrEnum):
    gpt_4o_mini = "gpt-4o-mini"
    gpt_5_mini = "gpt-5-mini"
    claude_sonnet_4_5 = "claude-sonnet-4-5"

    @property
    def provider(self) -> LLMProvider:
        return model_to_provider_map[self]

    @property
    def provider_string(self) -> str:
        return f"{self.provider.value}/{self.value}"

    @property
    def mode(self) -> instructor.Mode:
        return provider_to_mode_map[self.provider]


model_to_provider_map = {
    LLMModel.gpt_4o_mini: LLMProvider.OPENAI,
    LLMModel.gpt_5_mini: LLMProvider.OPENAI,
    LLMModel.claude_sonnet_4_5: LLMProvider.ANTHROPIC,
}

provider_to_mode_map = {
    LLMProvider.OPENAI: instructor.Mode.RESPONSES_TOOLS,
    LLMProvider.ANTHROPIC: instructor.Mode.ANTHROPIC_TOOLS,
}
